{ Containers }
unit Containr;

interface

uses
  Classes, SysUtils, World, Maths, ESensor;

type
  { Forward class declarations }
  TContainer = class;

  { TContainerSensor - TSensor subclass that can notify a container about its
    own destruction to avoid dangling pointers }
  TContainerSensor = class(TSensor)
  private
    // Owning container
    Container: TContainer;
  public
    // TSerializable methods
    function CanSerialize: Boolean; override;
  public
    destructor Destroy; override;
  end;

  { TContainer }
  TContainer = class(TStaticMeshEntity)
  private
    // Sensor entity, owned by the container
    Sensor: TContainerSensor;
    // Create the sensor
    procedure CreateSensor;
    // Destroy the sensor
    procedure DestroySensor;
    // Update the sensor's position
    procedure UpdateSensorPosition;
  protected
    // TSerializable methods
    procedure BeforeSerialization; override;
    procedure AfterSerialization; override;
    procedure TransformChanged; override;
  protected
    // TEntity methods
    procedure AddedToWorld; override;
    procedure RemovedFromWorld; override;
  end;

implementation

uses
  Serial;

{ TContainerSensor }
destructor TContainerSensor.Destroy;
begin
  Container.Sensor:=nil;
  inherited Destroy;
end;

function TContainerSensor.CanSerialize: Boolean;
begin
  Result:=False;
end;

{ TContainer }
procedure TContainer.CreateSensor;
begin
  // Destroy any existing sensor
  DestroySensor;
  // Create new sensor entity
  Sensor:=TContainerSensor.Create;
  Sensor.Container:=Self;
  World.AddEntity(Sensor);
  // Setup the target and command to be this container with the open command
  Sensor.Target:=Name;
  Sensor.Command:='open';
end;

procedure TContainer.DestroySensor;
begin
  Sensor.Free;
end;

procedure TContainer.UpdateSensorPosition;
var
  T: TTransform;
begin
  if Assigned(Sensor) and Assigned(Mesh) then begin
    T.Reset;
    T.Translation.Y:=0.9;
    T.Translation.Z:=Mesh.AABox.Max.Z + 0.1;
    Transform.ToMatrix.Transform(T.Translation);
    Sensor.Transform:=T;
  end;
end;

procedure TContainer.BeforeSerialization;
begin
  DestroySensor;
  inherited BeforeSerialization;
end;

procedure TContainer.AfterSerialization;
begin
  inherited AfterSerialization;
  CreateSensor;
  UpdateSensorPosition;
end;

procedure TContainer.TransformChanged;
begin
  inherited TransformChanged;
  UpdateSensorPosition;
end;

procedure TContainer.AddedToWorld;
begin
  inherited AddedToWorld;
  CreateSensor;
end;

procedure TContainer.RemovedFromWorld;
begin
  DestroySensor;
  inherited RemovedFromWorld;
end;

initialization
  RegisterSerializableClass(TContainer);
end.
