{ Doors }
unit EDoor;

interface

uses
  Classes, SysUtils, World, Maths, Scene, Meshes, Textures, Renderer;

type

  { TDoorState - Door states }
  TDoorState = (dsClosed, dsOpen, dsOpening, dsClosing);

  { TDoor - Door }
  TDoor = class(TSceneNodeEntity)
  private
    // Property storage
    FDoorHeight: Single;
    FTextureName: string;
    FDoorLength: Single;
    FState: TDoorState;
    FUpdatePortal: Boolean;
    procedure SetDoorHeight(ADoorHeight: Single);
    procedure SetDoorLength(ADoorLength: Single);
    procedure SetTextureName(ATextureName: string);
    procedure SetUpdatePortal(AUpdatePortal: Boolean);
  private
    // Door mesh
    Mesh: TMesh;
    // Y rotation when the state changed
    StartYRotation: Single;
    // Enable/disable any portals near the door during the next update
    UpdateNearbyPortalState: Boolean;
    // Update the door mesh
    procedure UpdateMesh;
  protected
    // TEntity methods
    procedure TransformChanged; override;
    procedure AddSceneNodes; override;
    procedure Update; override;
  protected
    // TSceneNodeEntity methods
    function CreateSceneNode: TSceneNode; override;
  public
    // TEntity methods
    function PerformCommand(ACommand: string): Boolean; override;
  public
    constructor Create; override;
    destructor Destroy; override;
    // Open the door
    procedure Open;
    // Close the door
    procedure Close;
    // Toggle the door's state
    procedure Toggle;
  published
    // Door state
    property State: TDoorState read FState;
    // Door height
    property DoorHeight: Single read FDoorHeight write SetDoorHeight;
    // Door length
    property DoorLength: Single read FDoorLength write SetDoorLength;
    // Door texture
    property TextureName: string read FTextureName write SetTextureName;
    // Update nearby portal based on door's state
    property UpdatePortal: Boolean read FUpdatePortal write SetUpdatePortal;
  end;

implementation

uses
  Engine, Serial;

{ TDoor }
constructor TDoor.Create;
begin
  inherited Create;
  FDoorHeight:=2;
  FDoorLength:=2;
  FTextureName:='door1.raw';
  FState:=dsClosed;
  FUpdatePortal:=True;
  Solid:=True;
  Mesh:=TMesh.Create;
  UpdateNearbyPortalState:=True;
  UpdateMesh;
end;

destructor TDoor.Destroy;
begin
  Mesh.Free;
  inherited Destroy;
end;

procedure TDoor.Open;
begin
  if State=dsClosed then begin
    FState:=dsOpening;
    StartYRotation:=RotationY;
    UpdateNearbyPortalState:=True;
  end;
end;

procedure TDoor.Close;
begin
  if State=dsOpen then begin
    FState:=dsClosing;
    StartYRotation:=RotationY;
  end;
end;

procedure TDoor.Toggle;
begin
  if State=dsOpen then Close else if State=dsClosed then Open;
end;

procedure TDoor.SetDoorHeight(ADoorHeight: Single);
begin
  if FDoorHeight=ADoorHeight then Exit;
  FDoorHeight:=ADoorHeight;
  UpdateMesh;
end;

procedure TDoor.SetDoorLength(ADoorLength: Single);
begin
  if FDoorLength=ADoorLength then Exit;
  FDoorLength:=ADoorLength;
  UpdateMesh;
end;

procedure TDoor.SetTextureName(ATextureName: string);
begin
  if FTextureName=ATextureName then Exit;
  FTextureName:=ATextureName;
  UpdateMesh;
end;

procedure TDoor.SetUpdatePortal(AUpdatePortal: Boolean);
begin
  if FUpdatePortal=AUpdatePortal then Exit;
  FUpdatePortal:=AUpdatePortal;
  if FUpdatePortal then UpdateNearbyPortalState:=True;
end;

procedure TDoor.UpdateMesh;
var
  Tri: TRenderTri;
  DoorMesh: TMesh;
  I: Integer;
begin
  Mesh.Clear;
  Mesh.Texture:=TTexture(GResourceManager.GetResource(TextureName));
  DoorMesh:=TMesh(GResourceManager.GetResource('doormesh.jtf'));
  for I:=0 to DoorMesh.TriangleCount - 1 do begin
    Tri:=DoorMesh.Triangles[I];
    Tri.A.Multiply(Vector(DoorLength, DoorHeight, 1));
    Tri.B.Multiply(Vector(DoorLength, DoorHeight, 1));
    Tri.C.Multiply(Vector(DoorLength, DoorHeight, 1));
    Mesh.AddTriangle(Tri);
  end;
  Mesh.UpdateAABox;
  if Assigned(SceneNode) then begin
    SceneNode.UpdateAABox;
    TMeshNode(SceneNode).Light:=World.GridMap.LightAt(Transform.Translation);
  end;
end;

procedure TDoor.TransformChanged;
begin
  inherited TransformChanged;
  UpdateMesh;
end;

procedure TDoor.AddSceneNodes;
begin
  inherited AddSceneNodes;
  UpdateMesh;
  UpdateNearbyPortalState:=True;
end;

procedure TDoor.Update;
var
  Portal: TScenePortal;
begin
  inherited Update;
  case State of
    dsOpening: begin
      if RotationY >= StartYRotation + 76 then begin
        FState:=dsOpen;
        UpdateNearbyPortalState:=True;
      end else
        RotationY:=RotationY + 2;
    end;
    dsClosing: begin
      if RotationY <= StartYRotation - 76 then begin
        FState:=dsClosed;
        UpdateNearbyPortalState:=True;
      end else
        RotationY:=RotationY - 2;
    end;
  end;
  if UpdatePortal and UpdateNearbyPortalState and Assigned(Scene) then begin
    Portal:=Scene.FindPortalNear(Transform.Translation, 2.5);
    if Assigned(Portal) then Portal.Enabled:=FState <> dsClosed;
    UpdateNearbyPortalState:=False;
  end;
end;

function TDoor.CreateSceneNode: TSceneNode;
begin
  Result:=TMeshNode.Create;
  TMeshNode(Result).Mesh:=Mesh;
  TMeshNode(Result).DepthBias:=12;
  UpdateMesh;
end;

function TDoor.PerformCommand(ACommand: string): Boolean;
begin
  // Ask parent to handle the command
  if inherited PerformCommand(ACommand) then Exit(True);
  // Check for door commands
  if ACommand='open' then begin
    if State <> dsClosed then Exit(False);
    Open;
    Exit(True);
  end;
  if ACommand='close' then begin
    if State <> dsOpen then Exit(False);
    Close;
    Exit(True);
  end;
  if ACommand='toggle' then begin
    if not (State in [dsOpen, dsClosed]) then Exit(False);
    Toggle;
    Exit(True);
  end;
  Result:=False;
end;

initialization
  RegisterSerializableClass(TDoor);
end.

