{ Petra entity }
unit EPetra;

interface

uses
  Classes, Maths, Serial, Engine, Scene, World, PieceMdl, InputDrv, ESensor,
  EPickup, Sounds, GridMap, Cameras;

type
  TPetraState = (psIdle, psRun, psPickUp, {psDoggify, psDogRun, psUndoggify,}
    psTurnLeft, psTurnRight, psRunTurnLeft, psRunTurnRight, psDogTurnLeft,
    psDogTurnRight, psFrontBlocked, psShortFall, psClimbMid, psClimbHigh,
    psJumpUp, psHanging, psSwingLeft, psSwingRight, psHangingClimb, psFalling,
    psJumping, psJumpingForward, psJumpingForwardFalling, psTurning,
    psSteppingBack, psActivating, psDying, psDead, psWaiting, psNoticed);

  { TPetra - entity for the player character }
  TPetra = class(TSceneNodeEntity, TInputHandler)
  private
    // Property storage
    FInventoryItems: array of string;
    function GetInventoryItems(AIndex: Integer): string; inline;
    function GetInventoryItemCount: Integer; inline;
  private
    // Current state
    State: TPetraState;
    // Petra piece model
    PetraModel: TPieceModel;
    // Petra's head scene node
    Head: TPieceModelPieceNode;
    // Petra's left arm scene node (used for pick up items)
    LeftArm: TPieceModelPieceNode;
    // Use for animation base
    Base: Integer;
    // Run speed
    RunSpeed: Single;
    // Fall speed (resets to 0 when not falling)
    FallSpeed: Single;
    // Animation blend factor
    AnimBlend: Single;
    // Translation Y component target (the real translation adjust itself
    // smoothly at every update to go to that value so that there are no
    // abrupt height changes)
    ToY: Single;
    // Amount to turn left/right next update (updated via Mouse)
    ToTurn: Single;
    // Current light levell
    CurrentLight: Single;
    // Target sensor for psActivating
    TargetSensor: TSensor;
    // Target pickup item for psPickUp
    TargetPickup: TPickupItem;
    // Transform to use after resurrection
    ResurrectionTransform: TTransform;
    // Step sound clip
    StepSound: TSoundClip;
    // If True, the step sound will be played at the end of the update cycle
    PlayStepSound: Boolean;
    // Hitting the ground sound clip
    HitGroundSound: TSoundClip;
    // Current camera specified by a grid
    GridCam: TCamera;
    // Set a new state, optionally reseting the animation base
    procedure SetState(NewState: TPetraState; ResetBase: Boolean);
    // Returns the floor height that would be if Petra moved at the given point
    // Optionally it also returns the ceiling height
    function GetFloorHeightAt(const P: TVector; ACeilingHeight: PSingle=nil): Single;
    // Resurrect after death
    procedure Resurrect;
  protected
    // TSerializable methods
    procedure AfterRootDeserialization; override;
  protected
    // TEntity methods
    procedure AddedToWorld; override;
    procedure RemovedFromWorld; override;
    procedure AddSceneNodes; override;
    procedure RemoveSceneNodes; override;
    procedure Update; override;
    procedure Frame; override;
  protected
    // TSceneNodeEntity methods
    function CreateSceneNode: TSceneNode; override;
  public
    // TEntity methods
    function PerformCommand(ACommand: string): Boolean; override;
  public
    // TInputHandler methods
    function KeyDown(Key: TInputKey): Boolean;
    function KeyUp(Key: TInputKey): Boolean;
    function Mouse(X, Y: Integer): Boolean;
  public
    // Create the entity
    constructor Create; override;
    // Destroy the entity
    destructor Destroy; override;
    // Add the given item to the inventory
    procedure AddToInventory(AItem: string);
    // Remove the given item from the inventory
    procedure RemoveFromInventory(AItem: string);
    // Returns the index of the given inventory item or -1 if the item is not
    // in the inventory (ignores case)
    function IndexOfInventoryItem(AItem: string): Integer;
    // Returns true if the given item is in the inventory
    function HasInventoryItem(AItem: string): Boolean; inline;
    // Apply the given skin file (empty string to restore the default skin)
    procedure ApplySkin(ASkin: string);
    // Returns true if Petra is dead
    function IsDead: Boolean; inline;
    // Inventory items
    property InventoryItems[AIndex: Integer]: string read GetInventoryItems;
    // Inventory item count
    property InventoryItemCount: Integer read GetInventoryItemCount;
  end;

implementation

uses
  SysUtils, DataLoad, Settings, Textures, Game, ScrEnt, GfxDrv;

var
  // Current gamma value
  Gamma: TGammaValue;

{ TPetra }
constructor TPetra.Create;
begin
  inherited Create;
  // Set the entity name
  Name:='Petra';
  // Load the piece model
  PetraModel:=TPieceModel.Create;
  PetraModel.LoadPMFModel(DataDirectoryPrefix + 'petra.pmf',
    DataDirectoryPrefix + 'petra.raw', 64, 64,
    0.45); // scale for ~1.82m height
  // Load sounds
  StepSound:=TSoundClip(GResourceManager.GetResource('step.wav'));
  HitGroundSound:=TSoundClip(GResourceManager.GetResource('hitdown.wav'));
  // Set a small height offset to avoid
  HeightOffset:=0.05;
  // Initial transform
  Transform:=Maths.Transform(Vector(0, 0, 0), Vector(0, -PI/2, 0));
end;

destructor TPetra.Destroy;
begin
  // Free the piece model
  PetraModel.Free;
  inherited Destroy;
end;

procedure TPetra.AddToInventory(AItem: string);
begin
  // Avoid duplicate inventory items
  if HasInventoryItem(AItem) then Exit;
  // Add the new item to the inventory
  SetLength(FInventoryItems, Length(FInventoryItems) + 1);
  FInventoryItems[High(FInventoryItems)]:=AItem;
end;

procedure TPetra.RemoveFromInventory(AItem: string);
var
  I, Index: Integer;
begin
  Index:=IndexOfInventoryItem(AItem);
  if Index=-1 then Exit;
  for I:=Index to High(FInventoryItems) - 1 do
    FInventoryItems[I]:=FInventoryItems[I + 1];
  SetLength(FInventoryItems, Length(FInventoryItems) - 1);
end;

function TPetra.IndexOfInventoryItem(AItem: string): Integer;
var
  I: Integer;
begin
  AItem:=LowerCase(AItem);
  for I:=0 to High(FInventoryItems) do
    if LowerCase(FInventoryItems[I])=AItem then Exit(I);
  Result:=-1;
end;

function TPetra.HasInventoryItem(AItem: string): Boolean;
begin
  Result:=IndexOfInventoryItem(AItem) <> -1;
end;

procedure TPetra.ApplySkin(ASkin: string);
begin
  if ASkin='' then
    PetraModel.OverrideTexture(nil)
  else
    PetraModel.OverrideTexture(TTexture(GResourceManager.GetResource(ASkin)));
end;

function TPetra.IsDead: Boolean;
begin
  Result:=State=psDead;
end;

function TPetra.GetInventoryItems(AIndex: Integer): string;
begin
  Result:=FInventoryItems[AIndex];
end;

function TPetra.GetInventoryItemCount: Integer;
begin
  Result:=Length(FInventoryItems);
end;

procedure TPetra.SetState(NewState: TPetraState; ResetBase: Boolean);
begin
  State:=NewState;
  if ResetBase then Base:=GGame.Milliseconds;
  AnimBlend:=0.25;
  if NewState=psFalling then FallSpeed:=0;
end;

function TPetra.GetFloorHeightAt(const P: TVector; ACeilingHeight: PSingle): Single;
var
  Box: TAABox;
  CeilingHeight: Single;
begin
  Box:=Maths.AABox(P.X - 0.25, P.Y, P.Z - 0.25,
                   P.X + 0.25, P.Y + 1.9, P.Z + 0.25);
  World.GetFloorAndCeilingHeightForBoxCorners(Box, Result, CeilingHeight);
  if Assigned(ACeilingHeight) then ACeilingHeight^:=CeilingHeight;
end;

procedure TPetra.Resurrect;
begin
  if State=psDead then begin
    SetState(psIdle, True);
    Transform:=ResurrectionTransform;
    ToY:=Transform.Translation.Y;
  end;
end;

procedure TPetra.AfterRootDeserialization;
begin
  inherited AfterRootDeserialization;
  // Store initial transform as resurrection transform
  ResurrectionTransform:=Transform;
end;

procedure TPetra.AddedToWorld;
begin
  inherited AddedToWorld;
  // Add Petra as an input handler
  GInput.AddHandler(Self);
  // Store current time as animation base
  Base:=GGame.Milliseconds;
  // Store initial ToY
  ToY:=Transform.Translation.Y;
  // Make Petra the player entity
  GGame.Player:=Self;
end;

procedure TPetra.RemovedFromWorld;
begin
  // Remove Petra from being the player entity
  if GGame.Player=Self then GGame.Player:=nil;
  // Remove Petra from being an input handler
  GInput.RemoveHandler(Self);
  inherited RemovedFromWorld;
end;

procedure TPetra.AddSceneNodes;
var
  PetraSkin, AmbientSound: string;
begin
  inherited AddSceneNodes;
  // Disable occlusion tests for the node to ensure Petra is always visible
  SceneNode.TestOcclusion:=False;
  // Setup initial camera position behind Petra's head so that it looks as if
  // it is zooming out from it
  with Scene.Camera do begin
    Position:=Head.WorldMatrix.Translation;
    Direction:=Head.WorldMatrix.TransformedWithoutTranslation(Vector(0, 0, 1));
  end;
  // Override skin if the world specifies a new one
  PetraSkin:=GetData('petraskin');
  if PetraSkin <> '' then
    PetraModel.OverrideTexture(TTexture(GResourceManager.GetResource(PetraSkin)));
  // Play ambient sound
  AmbientSound:=GetData('ambient');
  if AmbientSound <> '' then
    GSound.RepeatClip(TSoundClip(GResourceManager.GetResource(AmbientSound)), 10);
end;

procedure TPetra.RemoveSceneNodes;
var
  AmbientSound: string;
begin
  // Stop ambient sound
  AmbientSound:=GetData('ambient');
  if AmbientSound <> '' then
    GSound.StopClip(TSoundClip(GResourceManager.GetResource(AmbientSound)));
  inherited RemoveSceneNodes;
end;

procedure TPetra.Update;
var
  T: TTransform;
  UpdateTransform: Boolean = False;
  Pos, Dir: TVector;
  SaveRunSpeed, FloorHeight, CeilingHeight, Height, SaveToY: Single;
  TargetLight: Byte;
  I: Integer;
  MeshNode: TMeshNode;
  Mtx: TMatrix;
  Grid: TGrid;
  Cell: PCell;
  Entity: TEntity;

  // Return the floor height at the given point
  function GetPointFloorHeight(const P: TVector): Single;
  var
    A: Single;
  begin
    World.GetFloorAndCeilingHeightsAt(P, Result, A);
  end;

  // Check if Petra could fit in the given position
  function CanPetraFitInAt(X, Y, Z: Single): Boolean;
  var
    TestAA: TAABox;
  begin
    TestAA:=Maths.AABox(X - 0.25, Y, Z - 0.25, X + 0.25, Y + 1.9, Z + 0.25);
    Result:=World.CanBoxFit(TestAA);
  end;

  // Check if Petra's current position can fit in the world
  function CanPetraFitIn: Boolean; inline;
  begin
    Result:=CanPetraFitInAt(T.Translation.X, ToY, T.Translation.Z);
  end;

  // Make an attempt to move Petra at the given position
  procedure MoveTo(P: TVector);
  var
    TestAA: TAABox;
  begin
    P.Y:=ToY;
    TestAA:=Maths.AABox(P.X - 0.25, P.Y, P.Z - 0.25,
                        P.X + 0.25, P.Y + 1.9, P.Z + 0.25);
    World.GetFloorAndCeilingHeightForBoxCorners(TestAA, FloorHeight, CeilingHeight);
    if not World.CanBoxFit(TestAA) then begin
      if ((FloorHeight > P.Y) and (FloorHeight - P.Y < 0.7)) then begin
        if CanPetraFitInAt(P.X, FloorHeight, P.Z) then begin
          P.Y:=FloorHeight;
          UpdateTransform:=True;
        end;
      end;
    end else
      if FloorHeight < P.Y then begin
        if P.Y - FloorHeight > 0.8 then begin
          if P.Y - FloorHeight > 2 then begin
            SetState(psFalling, True);
            FallSpeed:=0.2;
          end else begin
            SetState(psShortFall, True);
            GSound.PlayClip(HitGroundSound, 32);
            P.Y:=FloorHeight;
          end;
        end else
          P.Y:=FloorHeight;
        UpdateTransform:=True;
      end;
    TestAA:=Maths.AABox(P.X - 0.25, P.Y, T.Translation.Z - 0.25,
                        P.X + 0.25, P.Y + 1.9, T.Translation.Z + 0.25);
    if World.CanBoxFit(TestAA) then begin
      T.Translation.X:=P.X;
      UpdateTransform:=True;
    end else
      RunSpeed:=RunSpeed*0.75;
    TestAA:=Maths.AABox(T.Translation.X - 0.25, P.Y, P.Z - 0.25,
                        T.Translation.X + 0.25, P.Y + 1.9, P.Z + 0.25);
    if World.CanBoxFit(TestAA) then begin
      T.Translation.Z:=P.Z;
      UpdateTransform:=True;
    end else
      RunSpeed:=RunSpeed*0.75;
    ToY:=P.Y;
  end;

begin
  // Ensure animation base is not in the future (can happen if the game is
  // initialized after the entity is added to the world)
  if Base > GGame.Milliseconds then Base:=GGame.Milliseconds;
  // Save running speed
  SaveRunSpeed:=RunSpeed;
  // Store transform to update later
  T:=Transform;
  SaveToY:=ToY;
  // Update rotation from mouse motion
  if ToTurn <> 0.0 then begin
    if State in [psIdle, psRun] then begin
      T.Rotation.Y += ToTurn;
      UpdateTransform:=True;
    end;
    ToTurn:=0.0;
  end;
  // Update rotation based on state
  case State of
    psTurnLeft, psRunTurnLeft, psDogTurnLeft: begin
      T.Rotation.Y += 0.045;
      UpdateTransform:=True;
    end;
    psTurnRight, psRunTurnRight, psDogTurnRight: begin
      T.Rotation.Y -= 0.045;
      UpdateTransform:=True;
    end;
  end;
  // Update motion based on state
  case State of
    psRun, psRunTurnLeft, psRunTurnRight: begin
      MoveTo(T.Translation.Added(Vector(
        Sin(T.Rotation.Y)*RunSpeed,
        0,
        Cos(T.Rotation.Y)*RunSpeed)));
      if Distance(T.Translation, Transform.Translation) < 0.01 then begin
        if RunSpeed > 0.04 then
          SetState(psFrontBlocked, True)
        else
          SetState(psIdle, True);
      end;
      if RunSpeed < 0.12 then RunSpeed:=RunSpeed + 0.01;
    end;
    {psDogRun, psDogTurnLeft, psDogTurnRight: begin
      MoveTo(T.Translation.Added(Vector(
        Sin(T.Rotation.Y)*0.05,
        0,
        Cos(T.Rotation.Y)*0.05)));
    end;}
    psFalling, psJumping, psJumpingForward, psJumpingForwardFalling: begin
      // If we're jumping forward, move forward
      if State=psJumpingForward then begin
        T.Translation.X += Sin(T.Rotation.Y)*0.18;
        T.Translation.Z += Cos(T.Rotation.Y)*0.18;
        UpdateTransform:=True;
        if not CanPetraFitIn then begin
          State:=psJumpingForwardFalling;
          FallSpeed *= 0.25;
          T:=Transform;
          // Go as close to collision point as possible so that grab will work
          repeat
            T.Translation.X += Sin(T.Rotation.Y)*0.018;
            T.Translation.Z += Cos(T.Rotation.Y)*0.018;
          until not CanPetraFitIn;
          T.Translation.X -= Sin(T.Rotation.Y)*0.018;
          T.Translation.Z -= Cos(T.Rotation.Y)*0.018;
        end;
      end;
      // Check the current floor height and see if we are on ground
      Height:=GetFloorHeightAt(T.Translation);
      if Height < ToY then begin
        ToY -= FallSpeed;
        FallSpeed += 0.01;
        if ToY <= Height then begin
          UpdateTransform:=True;
          if (State in [psJumpingForward, psJumpingForwardFalling]) or
             (FallSpeed > 0.1) then begin
            SetState(psShortFall, True);
            GSound.PlayClip(HitGroundSound, 32);
          end else
            SetState(psIdle, True);
          ToY:=Height;
        end;
      end else begin
        FallSpeed:=0;
        SetState(psIdle, True);
      end;
      // Check if we hit a wall or the ceiling
      if not CanPetraFitIn then begin
        // Restore previous position
        T:=Transform;
        ToY:=SaveToY;
        UpdateTransform:=False;
        // Stop jumping
        FallSpeed:=0;
      end;
      // If the grab key is pressed, check if we can grab on
      if GInput.KeyState[ikControl] or GInput.KeyState[ikLeftButton] then begin
        Height:=GetFloorHeightAt(Vector(
          T.Translation.X + Sin(T.Rotation.Y)*0.1,
          ToY + 1.97*0.45,
          T.Translation.Z + Cos(T.Rotation.Y)*0.1),
          @CeilingHeight);
        if {TODO: Fix (Abs(Height - CeilingHeight) > 0.01) and }(Abs(Height - (ToY + 1.845)) < 0.1) then begin
          if not CanPetraFitInAt(T.Translation.X, Height - 1.845, T.Translation.Z) then Exit;
          UpdateTransform:=True;
          ToY:=Height - 1.845;
          T.Translation.Y:=Interpolate(T.Translation.Y, ToY, 0.25);
          SetState(psHanging, True);
          PetraModel.ApplyAnimation(12, 0, TPieceModelNode(SceneNode), 1);
          T.Rotation.Y:=AxisAlignedAngleSnap(T.Rotation.Y);
        end;
      end;
    end;
  end;
  // Adjust the vertical translation to be towards ToY
  if Abs(ToY - T.Translation.Y) > 0.01 then begin
    if State in [psFalling, psJumping, psJumpingForward, psJumpingForwardFalling] then
      T.Translation.Y += (ToY - T.Translation.Y)*0.325
    else
      T.Translation.Y += (ToY - T.Translation.Y)*0.125;
    UpdateTransform:=True;
    if Abs(ToY - T.Translation.Y) <= 0.01 then T.Translation.Y:=ToY;
  end;

  // Calculate the final (non-interpolated) camera position based on the
  // current state
  if State in [psHanging, psSwingLeft, psSwingRight, psHangingClimb] then
    Pos:=SceneNode.WorldMatrix.Transformed(Vector(0, 3, -2.5))
  else if State=psTurning then
    Pos:=Head.WorldMatrix.Transformed(Vector(0, 0.5, -2.5))
  else
    Pos:=SceneNode.WorldMatrix.Transformed(Vector(0, 2.1, -2.5));

  // Interpolate the camera towards the final position
  Scene.Camera.Position:=Vector(
    Interpolate(Scene.Camera.Position.X, Pos.X, 0.1155),
    Interpolate(Scene.Camera.Position.Y, Pos.Y, 0.0555),
    Interpolate(Scene.Camera.Position.Z, Pos.Z, 0.1155));

  // Calculate the camera direction towards Petra's head
  Dir:=Direction(Scene.Camera.Position, Head.WorldMatrix.Translation);
  Dir.X:=Interpolate(Scene.Camera.Direction.X, Dir.X, 0.495);
  Dir.Y:=Interpolate(Scene.Camera.Direction.Y, Dir.Y, 0.095);
  Dir.Z:=Interpolate(Scene.Camera.Direction.Z, Dir.Z, 0.495);
  Scene.Camera.Direction:=Dir.Normalized;

  // Slow down if shift is pressed
  if GInput.KeyState[ikShift] then
    RunSpeed:=Min(0.03, RunSpeed);

  // Reset blending if the run speed changed to smooth out the animation
  // transitions
  if ((RunSpeed > 0.05) and (SaveRunSpeed <= 0.05)) or
     ((RunSpeed <= 0.05) and (SaveRunSpeed > 0.05)) then
    AnimBlend:=0.1;

  // Apply animations based on state (some also switch state)
  case State of
    psIdle, psTurnLeft, psTurnRight: begin
      PetraModel.ApplyAnimation(0, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
    end;
    psPickUp: begin
      I:=PetraModel.ApplyAnimation(1, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
      if (I >= 8) and Assigned(TargetPickup) then begin
        // Create temporary scene node for the item
        MeshNode:=TMeshNode.Create;
        MeshNode.Mesh:=TargetPickup.Mesh;
        MeshNode.Light:=LeftArm.Light;
        Mtx.Translation(0, -0.56, 0);
        MeshNode.LocalMatrix:=Mtx;
        LeftArm.AddChild(MeshNode);
        // Inform the item it was picked up
        TargetPickup.PickUp;
        TargetPickup:=nil; // Ensure PickUp wont be called more than once
      end else if I=-1 then begin
        // Delete temporary nodes
        LeftArm.DeleteAllChildren;
        // Reset state
        SetState(psIdle, True);
      end;
    end;
    {psDoggify: begin
      if PetraModel.ApplyAnimation(2, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then
        SetState(psDogRun, True);
    end;
    psUndoggify: begin
      if PetraModel.ApplyAnimation(4, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then
        SetState(psIdle, True);
    end;}
    psRun, psRunTurnLeft, psRunTurnRight: begin
      if RunSpeed > 0.05 then begin
        I:=PetraModel.ApplyAnimation(5, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
        if (I in [0, 1, 8, 9]) and (StepSound.UsedCount=0) then
          PlayStepSound:=True;
      end else
        PetraModel.ApplyAnimation(8, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
    end;
    {psDogRun, psDogTurnLeft, psDogTurnRight: begin
      PetraModel.ApplyAnimation(3, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
    end;}
    psFrontBlocked: begin
      if PetraModel.ApplyAnimation(7, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then
        SetState(psIdle, True);
    end;
    psShortFall: begin
      if PetraModel.ApplyAnimation(9, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then
        SetState(psIdle, True);
    end;
    psClimbMid: begin
      if PetraModel.ApplyAnimation(10, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X += Sin(T.Rotation.Y)*0.52*0.45;
        T.Translation.Y:=ToY + 1.97*0.45;
        T.Translation.Z += Cos(T.Rotation.Y)*0.52*0.45;
        ToY:=T.Translation.Y;
        AnimBlend:=1;
        UpdateTransform:=True;
        State:=psIdle;
        PetraModel.ApplyAnimation(0, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
      end;
    end;
    psClimbHigh: begin
      if PetraModel.ApplyAnimation(11, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X += Sin(T.Rotation.Y)*0.58*0.45;
        T.Translation.Y:=ToY + 2.92*0.45;
        T.Translation.Z += Cos(T.Rotation.Y)*0.58*0.45;
        ToY:=T.Translation.Y;
        AnimBlend:=1;
        UpdateTransform:=True;
        State:=psIdle;
        PetraModel.ApplyAnimation(0, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
      end;
    end;
    psJumpUp: begin
      if PetraModel.ApplyAnimation(13, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend) >= 6 then begin
        SetState(psJumping, True);
        PetraModel.ApplyAnimation(13, Base + 1000, TPieceModelNode(SceneNode), 1);
        FallSpeed:=-0.2;
        // Nudge the position upwards to ensure that state wont be reset to
        // nothing in the next cycle
        ToY += 0.01;
      end;
    end;
    psHanging: begin
      if GInput.KeyState[ikLeft] or GInput.KeyState[ikA] then begin
        if Abs(GetPointFloorHeight(T.Translation.Added(Vector(
          Sin(T.Rotation.Y + PI/2)*0.5 + Sin(T.Rotation.Y)*0.58,
          -T.Translation.Y + ToY + 4.13*0.45,
          Cos(T.Rotation.Y + PI/2)*0.5 + Cos(T.Rotation.Y)*0.58)))
          -
          GetPointFloorHeight(T.Translation.Added(Vector(
            Sin(T.Rotation.Y)*0.58,
            -T.Translation.Y + ToY + 4.13*0.45,
            Cos(T.Rotation.Y)*0.58)))) > 0.01 then Exit;
        SetState(psSwingLeft, True);
        Base:=GGame.Milliseconds;
      end else if GInput.KeyState[ikRight] or GInput.KeyState[ikD] then begin
        if Abs(GetPointFloorHeight(T.Translation.Added(Vector(
          Sin(T.Rotation.Y - PI/2)*0.5 + Sin(T.Rotation.Y)*0.58,
          -T.Translation.Y + ToY + 4.13*0.45,
          Cos(T.Rotation.Y - PI/2)*0.5 + Cos(T.Rotation.Y)*0.58)))
          -
          GetPointFloorHeight(T.Translation.Added(Vector(
            Sin(T.Rotation.Y)*0.58,
            -T.Translation.Y + ToY + 4.13*0.45,
            Cos(T.Rotation.Y)*0.58)))) > 0.01 then Exit;
        SetState(psSwingRight, True);
        Base:=GGame.Milliseconds;
      end else if GInput.KeyState[ikUp] or GInput.KeyState[ikW] then begin
        if CanPetraFitInAt(T.Translation.X + Sin(T.Rotation.Y)*0.58*0.45,
                           ToY + 4.13*0.45,
                           T.Translation.Z + Cos(T.Rotation.Y)*0.58*0.45) then begin
          SetState(psHangingClimb, True);
          Base:=GGame.Milliseconds;
        end;
      end;
    end;
    psSwingLeft: begin
      if PetraModel.ApplyAnimation(14, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X += Sin(T.Rotation.Y + PI/2)*0.4*0.45;
        T.Translation.Z += Cos(T.Rotation.Y + PI/2)*0.4*0.45;
        UpdateTransform:=True;
        SetState(psHanging, True);
        PetraModel.ApplyAnimation(12, 0, TPieceModelNode(SceneNode), 1);
      end;
    end;
    psSwingRight: begin
      if PetraModel.ApplyAnimation(15, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X += Sin(T.Rotation.Y - PI/2)*0.4*0.45;
        T.Translation.Z += Cos(T.Rotation.Y - PI/2)*0.4*0.45;
        UpdateTransform:=True;
        SetState(psHanging, True);
        PetraModel.ApplyAnimation(12, 0, TPieceModelNode(SceneNode), 1);
      end;
    end;
    psHangingClimb: begin
      if PetraModel.ApplyAnimation(16, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X += Sin(T.Rotation.Y)*0.58*0.45;
        T.Translation.Y:=ToY + 4.13*0.45; //4.13
        T.Translation.Z += Cos(T.Rotation.Y)*0.58*0.45;
        ToY:=T.Translation.Y;
        AnimBlend:=1;
        UpdateTransform:=True;
        State:=psIdle;
        PetraModel.ApplyAnimation(0, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
      end;
    end;
    psFalling, psJumping, psJumpingForwardFalling: begin
      if FallSpeed > 0.15 then
        PetraModel.ApplyAnimation(17, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), 0.1)
    end;
    psJumpingForward:
      PetraModel.ApplyAnimation(19, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), 1);
    psTurning: begin
      if PetraModel.ApplyAnimation(18, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Rotation.Y += PI;
        UpdateTransform:=True;
        SetState(psIdle, True);
        PetraModel.ApplyAnimation(0, 0, TPieceModelNode(SceneNode), 1);
      end;
    end;
    psSteppingBack: begin
      if PetraModel.ApplyAnimation(20, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then begin
        T.Translation.X -= Sin(T.Rotation.Y)*1*0.45;
        T.Translation.Z -= Cos(T.Rotation.Y)*1*0.45;
        UpdateTransform:=True;
        SetState(psFalling, True);
        PetraModel.ApplyAnimation(0, 0, TPieceModelNode(SceneNode), 1);
      end else begin
        if Abs(GetFloorHeightAt(Head.WorldMatrix.Translation) - ToY) > 0.1 then begin
          T.Translation.X:=Head.WorldMatrix.Translation.X;
          T.Translation.Z:=Head.WorldMatrix.Translation.Z;
          UpdateTransform:=True;
          SetState(psFalling, True);
          PetraModel.ApplyAnimation(17, 0, TPieceModelNode(SceneNode), 0.4);
        end;
      end;
    end;
    psActivating: begin
      I:=PetraModel.ApplyAnimation(21, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
      if I >= 7 then begin
        if Assigned(TargetSensor) then begin
          TargetSensor.Activate;
          TargetSensor:=nil;
        end;
      end;
      if I=-1 then SetState(psIdle, True);
    end;
    psDying: begin
      if PetraModel.ApplyAnimation(22, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend)=-1 then
        SetState(psDead, True);
    end;
    psWaiting:
      PetraModel.ApplyAnimation(23, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
    psNoticed:
      PetraModel.ApplyAnimation(24, (GGame.Milliseconds - Base)/1000, TPieceModelNode(SceneNode), AnimBlend);
  end;

  // Update transform
  if UpdateTransform then begin
    Transform:=T;
    if not (State in [psDying, psDead]) then begin
      // Check the grid we're on
      Grid:=World.GridMap.GridAt(T.Translation);
      if Assigned(Grid) then begin
        // If the grid has a camera, check if we need to use it
        if Grid.Camera <> GridCam then begin
          if Assigned(Grid.Camera) then
            GGame.CameraOverride:=Grid.Camera
          else
            GGame.CameraOverride:=nil;
          GridCam:=Grid.Camera;
        end;
        // If there is a grid camera, update its direction
        if Assigned(GridCam) then begin
          GridCam.Direction:=Interpolate(GridCam.Direction,
            Direction(GridCam.Position, T.Translation),
            0.15).Normalized;
        end;
        // Check the cell
        Cell:=Grid.GetCellAt(T.Translation.X, T.Translation.Z);
        if Assigned(Cell) then begin
          // Check if we stepped on fatal floor
          if coFatalFloor in Cell^.CellOptions then begin
            World.GridMap.GetFloorAndCeilingHeightsAt(T.Translation, FloorHeight, CeilingHeight);
            if Abs(FloorHeight - T.Translation.Y) < 0.15 then begin
              ToY:=FloorHeight;
              SetState(psDying, True);
            end;
          end;
          // Check if we stepped on a script floor
          if coRunScript in Cell^.CellOptions then begin
            Entity:=World.FindClosestEntity(T.Translation, TScriptEntity, 2);
            if Assigned(Entity) then TScriptEntity(Entity).Run;
          end;
        end;
      end;
    end;
  end;

  // Update lighting
  TargetLight:=Round(Clamp(World.GridMap.LightAt(T.Translation), 0.0, 255.0/World.Overbright));
  CurrentLight:=Interpolate(CurrentLight, TargetLight, 0.15);
  if Round(CurrentLight) <> Head.Light then
    TPieceModelNode(SceneNode).Root.ApplyLightLevel(Round(CurrentLight));

  // Increase the animation blending factor, we only want to cross-blend
  // animations for a little while
  AnimBlend:=Min(1, AnimBlend + 0.1);
  
  // If we hit a step frame during animation, play the step sound at the end of
  // the update cycle (avoids multiple sounds with low framerates in case we
  // hit a step frame more than once during the update)
  if PlayStepSound and GFinalUpdateCall then begin
    GSound.PlayClip(StepSound, 32);
    PlayStepSound:=False;
  end;

  inherited Update;
end;

procedure TPetra.Frame;
begin
end;

function TPetra.CreateSceneNode: TSceneNode;
begin
  // Create a scene node for the Petra model
  Result:=PetraModel.CreateSceneNode;
  // Set a depth bias so that Petra is drawn a bit above the rest of polygons
  TPieceModelNode(Result).Root.ApplyDepthBias(14);
  // Find the head and arm nodes in the model
  Head:=PetraModel.GetPieceSceneNode('head', TPieceModelNode(Result));
  LeftArm:=PetraModel.GetPieceSceneNode('larml', TPieceModelNode(Result));
  // Initial light level
  TPieceModelNode(Result).Root.ApplyLightLevel(255);
  CurrentLight:=255;
end;

function TPetra.PerformCommand(ACommand: string): Boolean;
begin
  if inherited PerformCommand(ACommand) then Exit(True);
  if Copy(ACommand, 1, 8)='additem:' then begin
    AddToInventory(Copy(ACommand, 9, MaxInt));
    Exit(True);
  end;
  if Copy(ACommand, 1, 8)='remitem:' then begin
    RemoveFromInventory(Copy(ACommand, 9, MaxInt));
    Exit(True);
  end;
  if ACommand='toytotransform' then begin
    ToY:=Transform.Translation.Y;
    Exit(True);
  end;
  if ACommand='notereader' then begin
    GScreenManager.Screen:=GGame.NoteReaderScreen;
    Exit(True);
  end;
  if ACommand='die' then begin
    if not (State in [psDying, psDead]) then
      SetState(psDying, True);
    Exit(True);
  end;
  if ACommand='gowait' then begin
    SetState(psWaiting, True);
    Exit(True);
  end;
  if ACommand='gonotice' then begin
    SetState(psNoticed, True);
    Exit(True);
  end;
  Result:=False;
end;

function TPetra.KeyDown(Key: TInputKey): Boolean;
var
  P: TVector;
  Height: Single;
  TestAA: TAABox;

  // Check if Petra could fit in the given position
  function CanPetraFitInAt(X, Y, Z: Single): Boolean;
  var
    TestAA: TAABox;
  begin
    TestAA:=Maths.AABox(X - 0.25, Y, Z - 0.25, X + 0.25, Y + 1.9, Z + 0.25);
    Result:=World.CanBoxFit(TestAA);
  end;

begin
  Result:=True;
  case Key of
    ikEscape: GScreenManager.Screen:=TGame(GGame).InventoryScreen;
    ikShift: AnimBlend:=0;
    ikUp, ikW: begin
      case State of
        psIdle: SetState(psRun, False);
        psTurnLeft: SetState(psRunTurnLeft, False);
        psTurnRight: SetState(psRunTurnRight, False);
      end;
      RunSpeed:=0.02;
    end;
    ikLeft, ikA: case State of
      psIdle: SetState(psTurnLeft, False);
      psRun: SetState(psRunTurnLeft, False);
      //psDogRun: SetState(psDogTurnLeft, False);
    end;
    ikRight, ikD: case State of
      psIdle: SetState(psTurnRight, False);
      psRun: SetState(psRunTurnRight, False);
      //psDogRun: SetState(psDogTurnRight, False);
    end;
    ikDown, ikS: if State=psIdle then begin
      P.X:=Transform.Translation.X - Sin(Transform.Rotation.Y)*0.45;
      P.Z:=Transform.Translation.Z - Cos(Transform.Rotation.Y)*0.45;
      TestAA:=Maths.AABox(P.X - 0.25, ToY, P.Z - 0.25,
                          P.X + 0.25, ToY + 1.9, P.Z + 0.25);
      if World.CanBoxFit(TestAA) then SetState(psSteppingBack, True);
    end;
    ikControl, ikLeftButton: begin
      // Check for sensor
      if State=psIdle then begin
        P:=Head.WorldMatrix.Translation;
        P.X += Sin(Transform.Rotation.Y)*0.25;
        P.Y -= 0.25;
        P.Z += Cos(Transform.Rotation.Y)*0.25;
        TargetSensor:=TSensor(World.FindClosestEntity(P, TSensor, 0.3));
        if Assigned(TargetSensor) and (TargetSensor.State <> ssDisabled) then
          SetState(psActivating, True)
        else
          TargetSensor:=nil;
      end;
      // Check for jump or grab
      if State in [psIdle, psRun, psRunTurnLeft, psRunTurnRight] then begin
        P:=Vector(Transform.Translation.X + Sin(Transform.Rotation.Y)*0.52*0.45,
                  ToY + 1.97*0.45,
                  Transform.Translation.Z + Cos(Transform.Rotation.Y)*0.52*0.45);
        Height:=GetFloorHeightAt(P);
        if Height - 0.5 > ToY then begin
          if Height - ToY < 1.1 then begin
            if not CanPetraFitInAt(
              Transform.Translation.X + Sin(Transform.Rotation.Y)*0.52*0.45,
              Height,
              Transform.Translation.Z + Cos(Transform.Rotation.Y)*0.52*0.45) then Exit;
            ToY:=Height - 1.97*0.45;
            SetState(psClimbMid, True);
          end else if Height - ToY < 2 then begin
            if not CanPetraFitInAt(
              Transform.Translation.X + Sin(Transform.Rotation.Y)*0.58*0.45,
              Height,
              Transform.Translation.Z + Cos(Transform.Rotation.Y)*0.58*0.45) then Exit;
            ToY:=Height - 2.92*0.45;
            SetState(psClimbHigh, True);
          end else begin
            P:=Vector(Transform.Translation.X + Sin(AxisAlignedAngleSnap(Transform.Rotation.Y))*0.52*0.45,
                      ToY + 1.97*0.45,
                      Transform.Translation.Z + Cos(AxisAlignedAngleSnap(Transform.Rotation.Y))*0.52*0.45);
            if Height=GetFloorHeightAt(P) then begin
              Transform:=Maths.Transform(Transform.Translation,
                Vector(0, AxisAlignedAngleSnap(Transform.Rotation.Y), 0));
              SetState(psJumpUp, True);
            end;
          end;
        end;
      end;
      {$IFDEF ANDROID}
      if State=psDead then Resurrect;
      {$ENDIF}
    end;
    ikAlt, ikRightButton: case State of
      psIdle: SetState(psJumpUp, True);
      psRun, psRunTurnLeft, psRunTurnRight: begin
        SetState(psJumpingForward, True);
        FallSpeed:=-0.17;
        // Nudge the position upwards to ensure that state wont be reset to
        // nothing in the next cycle
        ToY += 0.01;
      end;
      psHanging: SetState(psFalling, True);
      {$IFDEF ANDROID}
      psDead: Resurrect;
      {$ENDIF}
    end;
    ikSpace: case State of
      //psIdle, psRun: SetState(psDoggify, True);
      //psDogRun: SetState(psUndoggify, True);
      psDead: Resurrect;
    end;
    ikEnter, ikMiddleButton: if State=psIdle then begin
      // Check if there is a pickup item at our feet
      TargetPickup:=TPickupItem(World.FindClosestEntity(Transform.Translation, TPickupItem, 0.3));
      if Assigned(TargetPickup) then SetState(psPickUp, True);
    end
    {$IFDEF ANDROID}
    else if State=psDead then Resurrect
    {$ENDIF}
    ;
    ikEnd, ikR: if State=psIdle then SetState(psTurning, True);
    ikF2: begin
      if Gamma=Low(TGammaValue) then Gamma:=High(TGammaValue) else Dec(Gamma);
      GRenderer.GraphicsDriver.SetGamma(Gamma);
    end;
    ikF3: begin
      if Gamma=High(TGammaValue) then Gamma:=Low(TGammaValue) else Inc(Gamma);
      GRenderer.GraphicsDriver.SetGamma(Gamma);
    end;
    else Result:=False;
  end;
end;

function TPetra.KeyUp(Key: TInputKey): Boolean;
begin
  Result:=True;
  case Key of
    ikUp, ikW: case State of
      psRun: SetState(psIdle, False);
      psRunTurnLeft: SetState(psTurnLeft, False);
      psRunTurnRight: SetState(psTurnRight, False);
    end;
    ikLeft, ikA: case State of
      psTurnLeft: SetState(psIdle, False);
      psRunTurnLeft: SetState(psRun, False);
      //psDogTurnLeft: SetState(psDogRun, False);
    end;
    ikRight, ikD: case State of
      psTurnRight: SetState(psIdle, False);
      psRunTurnRight: SetState(psRun, False);
      //psDogTurnRight: SetState(psDogRun, False);
    end;
    else Result:=False;
  end;
end;

function TPetra.Mouse(X, Y: Integer): Boolean;
{$push}{$warn 5024 off}
begin
  if X <> 0 then ToTurn -= X*0.003*(GSettings.MouseRotateSpeed*0.01);
  Result:=False;
end;
{$pop}

initialization
  // Register the Petra entity class
  RegisterSerializableClass(TPetra);
end.

