{ Sensors }
unit ESensor;

interface

uses
  Classes, SysUtils, Maths, World;

type

  { TSensorState - Sensor states }
  TSensorState = (ssEnabled, ssDisabled, ssCamera);

  { TSensor - Action sensor }
  TSensor = class(TEntity)
  private
    // Property storage
    FTarget: string;
    FCommand: string;
    FCamera: string;
    FCameraDuration: Word;
    FState: TSensorState;
    FItemNeeded: string;
    FItemMissingSpeech: string;
    FActivationSpeech: string;
    FActivationSound: string;
  private
    // The time stamp to release the camera
    CameraSwitchBackTime: Integer;
  protected
    // TEntity methods
    procedure Update; override;
    {$IFDEF EDITOR}
    procedure Frame; override;
    {$ENDIF}
    function GetAABox: TAABox; override;
  public
    // TEntity methods
    function PerformCommand(ACommand: string): Boolean; override;
  public
    constructor Create; override;
    // Activate the sensor
    procedure Activate;
  published
    // Sensor state
    property State: TSensorState read FState write FState;
    // Sensor target
    property Target: string read FTarget write FTarget;
    // Command to send to the target when the sensor is activated
    property Command: string read FCommand write FCommand;
    // Camera to switch to when the command is performed
    property Camera: string read FCamera write FCamera;
    // Camera switch duration in milliseconds
    property CameraDuration: Word read FCameraDuration write FCameraDuration;
    // Optional necessary item to activate the sensor
    property ItemNeeded: string read FItemNeeded write FItemNeeded;
    // Speech for missing item
    property ItemMissingSpeech: string read FItemMissingSpeech write FItemMissingSpeech;
    // Speech for activation
    property ActivationSpeech: string read FActivationSpeech write FActivationSpeech;
    // Sound for activation
    property ActivationSound: string read FActivationSound write FActivationSound;
  end;
  
implementation

uses
  Serial, Engine, CamEnt, EPetra, SpchScn, Sounds;

{ TSensor }
constructor TSensor.Create;
begin
  inherited Create;
  FCameraDuration:=1500;
end;

procedure TSensor.Update;
begin
  if State=ssCamera then begin
    if GGame.Milliseconds >= CameraSwitchBackTime then begin
      // Remove the override
      GGame.CameraOverride:=nil;
      FState:=ssEnabled;
    end;
  end;
end;

{$IFDEF EDITOR}
procedure TSensor.Frame;
begin
  if GPlayMode=pmEditor then GRenderer.RenderAABox(AABox);
end;
{$ENDIF}

function TSensor.GetAABox: TAABox;
begin
  Result:=inherited GetAABox;
  Result.Grow(0.1);
  Result.Min.Y -= 0.1;
  Result.Max.Y += 0.1;
end;

function TSensor.PerformCommand(ACommand: string): Boolean;
begin
  // Ask parent to handle the command
  if inherited PerformCommand(ACommand) then Exit(True);
  // Check for sensor commands
  if ACommand='activate' then begin
    if State <> ssEnabled then Exit(False);
    Activate;
    Exit(True);
  end;
  if ACommand='disable' then begin
    if State=ssDisabled then Exit(False);
    FState:=ssDisabled;
    Exit(True);
  end;
  if ACommand='enable' then begin
    if State <> ssDisabled then Exit(False);
    FState:=ssEnabled;
    Exit(True);
  end;
  Result:=False;
end;

procedure TSensor.Activate;
var
  Entity: TEntity;
  CamEntity: TCameraEntity;
  Clip: TSoundClip;
begin
  // Ensure we're enabled
  if State <> ssEnabled then Exit;
  // Check if we need an item
  if ItemNeeded <> '' then begin
    if not TPetra(GGame.Player).HasInventoryItem(ItemNeeded) then begin
      if ItemMissingSpeech='' then
        GGame.SpeechScreen.Speech(sstPetra, 'Something is missing here...', '', '')
      else
        GGame.SpeechScreen.Speech(sstPetra, ItemMissingSpeech, '', '');
      Exit;
    end;
  end;
  // Check if we should show some speech
  if ActivationSpeech <> '' then
    GGame.SpeechScreen.Speech(sstPetra, ActivationSpeech, '', '');
  // Find the entity
  Entity:=World.FindEntity(Target);
  if Assigned(Entity) then begin
    // Ask to perform the command
    if Entity.PerformCommand(Command) then begin
      // If we have a sound, play it
      if ActivationSound <> '' then begin
        Clip:=TSoundClip(GResourceManager.GetResource(ActivationSound));
        if Assigned(Clip) then GSound.PlayClip(Clip);
      end;
      // If we have a camera, switch to it
      if Camera <> '' then begin
        CamEntity:=TCameraEntity(World.FindEntity(Camera));
        if Assigned(CamEntity) then begin
          GGame.CameraOverride:=CamEntity.Camera;
          CameraSwitchBackTime:=GGame.Milliseconds + CameraDuration;
          FState:=ssCamera;
        end;
      end;
    end;
  end;
end;

initialization
  RegisterSerializableClass(TSensor);
end.
