{ Game side for the game }
unit Game;

interface

uses
  Misc, Serial, GameBase, Scripts, Cameras, InputDrv, InvenScn, CodeScn,
  ReadrScn, SpchScn, GOverScn;

type
  { TGame - Main game class }
  TGame = class(TGameBase,TInputHandler)
  private
    // Property storage
    FInventoryScreen: TInventoryScreen;
    FCodeScreen: TCodeScreen;
    FNoteReaderScreen: TNoteReaderScreen;
    FSpeechScreen: TSpeechScreen;
    FGameOverScreen: TGameOverScreen;
  private
    // Benchmark camera
    BenchCam: TCamera;
  protected
    // TGameBase methods
    procedure InitGame; override;
  public
    // TGameBase methods
    function RunScriptCommand(Script: TScript; Args: TStringArray): Boolean; override;
    procedure Update; override;
    procedure Frame; override;
    procedure PostFrame; override;
  public
    // TInputHandler methods
    function KeyDown(Key: TInputKey): Boolean;
    function KeyUp(Key: TInputKey): Boolean;
    function Mouse(X, Y: Integer): Boolean;
  public
    constructor Create; override;
    destructor Destroy; override;
    // Inventory screen
    property InventoryScreen: TInventoryScreen read FInventoryScreen;
    // Code screen
    property CodeScreen: TCodeScreen read FCodeScreen;
    // Note reader screen
    property NoteReaderScreen: TNoteReaderScreen read FNoteReaderScreen;
    // Speech screen
    property SpeechScreen: TSpeechScreen read FSpeechScreen;
    // Game over screen
    property GameOverScreen: TGameOverScreen read FGameOverScreen;
  end;
  
implementation

uses
  { Entity classes }
  EPetra, EDoor, ESensor, ERotor, EPickup, Containr,
  { Other units }
  Classes, Math, Maths, Engine, Renderer, World, GridMap, GfxDrv,
  Fonts, Settings, DataLoad;

{ TGame }
constructor TGame.Create;
begin
  inherited Create;
  FInventoryScreen:=TInventoryScreen.Create;
  FCodeScreen:=TCodeScreen.Create;
  FNoteReaderScreen:=TNoteReaderScreen.Create;
  FSpeechScreen:=TSpeechScreen.Create;
  FGameOverScreen:=TGameOverScreen.Create;
  if GBenchmark then BenchCam:=TCamera.Create;
end;

destructor TGame.Destroy;
begin
  BenchCam.Free;
  FGameOverScreen.Free;
  FSpeechScreen.Free;
  FNoteReaderScreen.Free;
  FCodeScreen.Free;
  FInventoryScreen.Free;
  inherited Destroy;
end;

procedure TGame.InitGame;
{$IFNDEF EDITOR}
var
  InitialMap: string;
  I: Integer;
{$ENDIF}
begin
  // Load game color palette
  {$IFDEF EDITOR}
  GRenderer.ColorPalette.LoadJASCPalette(DataDirectoryPrefix + 'colors.pal');
  GRenderer.ColorPalette.BuildLightTable(0, 0, 0, 1, Vector(1, 1, 1));
  {$ENDIF}
  {$IFNDEF EDITOR}
  InitialMap:='map1';
  if not GBenchmark then begin
    // Check if a custom map is specified via the settings
    if GSettings.CustomMap <> '' then InitialMap:=GSettings.CustomMap;
    // Check if a custom map is specified in command line
    for I:=1 to ParamCount do
      if (I < ParamCount) and (LowerCase(ParamStr(I))='-map') then
        InitialMap:=ParamStr(I + 1);
  end else begin
    GInput.ExclusiveHandler:=Self;
  end;
  PlayMap(InitialMap);
  {$ENDIF}
end;

function TGame.RunScriptCommand(Script: TScript; Args: TStringArray): Boolean;
var
  Theme: TSpeechScreenTheme;
  I: Integer;
begin
  if inherited RunScriptCommand(Script, Args) then Exit(True);
  // Command: say who text [target command]
  if Args[0]='say' then begin
    if Length(Args) < 3 then Exit(True);
    Theme:=sstPetra;
    if Args[1]='computer' then Theme:=sstComputer;
    if Args[1]='helen' then Theme:=sstHelen;
    if Length(Args) > 4 then
      SpeechScreen.Speech(Theme, Args[2], Args[3], Args[4])
    else
      SpeechScreen.Speech(Theme, Args[2], '', '');
    Exit(True);
  end;
  // Command: blankspeech yes/no
  if Args[0]='blankspeech' then begin
    if Length(Args) < 2 then Exit(True);
    SpeechScreen.BlankOut:=LowerCase(Args[1])='yes';
    Exit(True);
  end;
  // Command: reskin skinfile
  if Args[0]='reskin' then begin
    if Length(Args) > 1 then
      TPetra(Player).ApplySkin(Args[1])
    else
      TPetra(Player).ApplySkin('');
    Exit(True);
  end;
  // Command: additem itemname
  if Args[0]='additem' then begin
    if Length(Args) > 1 then
      TPetra(Player).AddToInventory(Args[1]);
    Exit(True);
  end;
  // Command: remitem itemname
  if Args[0]='remitem' then begin
    if Length(Args) > 1 then
      TPetra(Player).RemoveFromInventory(Args[1]);
    Exit(True);
  end;
  // Command: hasitem varname itemname
  if Args[0]='hasitem' then begin
    if Length(Args) < 3 then Exit(True);
    if TPetra(Player).HasInventoryItem(Args[2]) then
      Script.SetVar(Args[1], '1')
    else
      Script.SetVar(Args[1], '0');
    Exit(True);
  end;
  // Command: askcode code target command
  if Args[0]='askcode' then begin
    if Length(Args) < 4 then Exit(True);
    CodeScreen.ShowForCode(Args[1], Args[2], Args[3]);
    Exit(True);
  end;
  // Command: startallrotors
  if Args[0]='startallrotors' then begin
    for I:=0 to World.EntityCount - 1 do
      if World.Entities[I] is TRotor then
        TRotor(World.Entities[I]).State:=rsRotating;
    Exit(True);
  end;
  // Command: gameover
  if Args[0]='gameover' then begin
    GScreenManager.Screen:=GameOverScreen;
    Exit(True);
  end;
  Result:=False;
end;

procedure TGame.Update;
const
  SequenceData: array [0..371] of Single = (
    16.98, 3.59, 6.88, -1.00, 0.04, 0.01,
    12.59, 3.77, 6.92, -1.00, 0.04, 0.01,
    4.35, 3.13, 6.99, -1.00, 0.04, 0.01,
    -0.65, 3.38, 5.82, -0.44, 0.05, -0.89,
    0.73, 3.18, 4.30, 0.46, -0.08, -0.88,
    4.37, 2.75, 1.76, -0.30, -0.26, -0.92,
    3.12, 1.48, -1.70, -0.27, -0.02, -0.96,
    0.76, 1.48, -1.29, -0.02, 0.00, -1.00,
    -1.16, 2.73, -1.70, 0.34, -0.30, -0.89,
    4.09, 2.64, 0.23, -0.56, -0.33, -0.76,
    0.19, 1.99, -0.57, -0.99, -0.11, -0.07,
    -5.01, 1.52, -0.41, -0.29, -0.06, 0.96,
    -5.82, 1.50, 3.08, 0.11, -0.24, 0.96,
    -4.42, 2.81, 3.66, -0.89, -0.45, 0.09,
    -5.96, 2.81, 6.29, -0.37, -0.42, -0.83,
    -5.31, 1.81, 0.43, -0.99, -0.11, 0.03,
    -3.40, 2.02, 0.37, -0.99, -0.11, 0.03,
    -16.79, 1.84, 1.42, -1.00, -0.01, 0.09,
    -19.16, 2.92, -0.64, -0.05, -0.23, 0.97,
    -13.24, 3.60, 3.30, -0.60, -0.07, 0.80,
    -16.48, 3.60, 2.79, 0.37, -0.04, 0.93,
    -13.46, 2.38, -1.67, -0.36, -0.12, 0.93,
    -11.44, 2.37, 2.92, -0.97, -0.09, -0.21,
    -16.34, 1.99, 3.04, -0.99, -0.05, -0.10,
    -24.47, 2.03, 5.00, -0.31, -0.01, -0.95,
    -25.08, 2.25, -0.40, -0.06, 0.05, -1.00,
    -25.45, 2.74, -5.75, -0.88, -0.01, -0.48,
    -26.13, 2.64, -7.96, -0.96, -0.15, 0.24,
    -34.37, 1.82, -7.08, -1.00, -0.05, 0.03,
    -36.87, 1.67, -9.26, -0.23, -0.02, -0.97,
    -36.86, 1.63, -17.07, 0.04, 0.06, -1.00,
    -35.51, 3.18, -21.67, -1.00, 0.07, 0.03,
    -37.47, 1.58, -21.49, -0.95, 0.30, 0.05,
    -37.50, 3.68, -19.52, 0.02, -0.23, 0.97,
    -36.78, 1.93, -8.63, 0.04, -0.02, 1.00,
    -37.42, 2.00, -5.48, 0.98, -0.08, -0.16,
    -34.21, 2.23, -6.12, 0.98, 0.07, -0.20,
    -27.22, 2.91, -6.26, 0.94, -0.04, 0.34,
    -25.80, 2.79, -3.20, 0.12, -0.15, 0.98,
    -24.37, 1.77, 3.76, 0.38, -0.14, 0.91,
    -21.94, 1.77, 3.38, -0.28, -0.18, 0.94,
    -22.81, 1.29, 28.09, -0.05, 0.01, 1.00,
    -23.14, 1.35, 34.40, -0.03, 0.01, 1.00,
    -21.35, 1.34, 34.39, -0.45, 0.03, 0.89,
    -24.02, 1.68, 29.90, 0.17, -0.14, 0.98,
    -24.02, 1.68, 29.90, 0.87, 0.02, -0.49,
    -22.41, 2.36, 24.62, 0.03, -0.02, -1.00,
    -22.50, 1.48, 14.03, -0.03, -0.06, -1.00,
    -22.61, 1.26, 4.18, 0.76, -0.08, -0.64,
    -18.71, 1.73, 1.82, 0.95, -0.30, -0.07,
    -16.14, 1.75, 0.14, 0.27, -0.22, 0.94,
    -14.19, 1.75, 0.24, -0.35, -0.22, 0.91,
    -14.86, 1.64, 2.56, 0.85, -0.13, -0.51,
    -10.48, 1.22, 0.97, 1.00, -0.04, 0.01,
    -1.17, 1.44, 0.21, 0.92, -0.06, -0.38,
    2.82, 1.44, 0.71, -0.19, -0.02, -0.98,
    0.68, 1.50, -0.68, 0.11, 0.04, -0.99,
    1.62, 2.70, -0.99, 0.66, -0.33, 0.67,
    1.62, 2.70, -0.99, 0.22, -0.64, 0.73,
    -0.16, 2.67, -1.15, 0.16, 0.01, 0.99,
    1.71, 2.83, 0.39, -0.21, 0.07, 0.98,
    0.99, 2.83, 0.34, 0.00, 0.00, 1.00);
  BenchBase: Single = 0;
var
  SeqRow, NextRow, PrevRow, NextNextRow: Integer;
  BenchTicks, Interp: Single;
begin
  inherited Update;
  if GBenchmark then begin
    if BenchBase < 0.00001 then BenchBase:=GSystem.Milliseconds;
    BenchTicks:=GSystem.Milliseconds - BenchBase;
    SeqRow:=Clamp(Floor(BenchTicks/1000), 0, 61);
    if SeqRow=61 then QuitEngine;
    PrevRow:=Clamp(SeqRow -1, 0, 61);
    NextRow:=Clamp(SeqRow + 1, 0, 61);
    NextNextRow:=Clamp(SeqRow + 2, 0, 61);
    Interp:=BenchTicks/1000 - SeqRow;
    GRenderer.Camera.Position:=CatmullRomInterpolate(
      Vector(SequenceData[PrevRow*6], SequenceData[PrevRow*6 + 1], SequenceData[PrevRow*6 + 2]),
      Vector(SequenceData[SeqRow*6], SequenceData[SeqRow*6 + 1], SequenceData[SeqRow*6 + 2]),
      Vector(SequenceData[NextRow*6], SequenceData[NextRow*6 + 1], SequenceData[NextRow*6 + 2]),
      Vector(SequenceData[NextNextRow*6], SequenceData[NextNextRow*6 + 1], SequenceData[NextNextRow*6 + 2]),
      Interp);
    GRenderer.Camera.Direction:=CatmullRomInterpolate(
      Vector(SequenceData[PrevRow*6 + 3], SequenceData[PrevRow*6 + 4], SequenceData[PrevRow*6 + 5]),
      Vector(SequenceData[SeqRow*6 + 3], SequenceData[SeqRow*6 + 4], SequenceData[SeqRow*6 + 5]),
      Vector(SequenceData[NextRow*6 + 3], SequenceData[NextRow*6 + 4], SequenceData[NextRow*6 + 5]),
      Vector(SequenceData[NextNextRow*6 + 3], SequenceData[NextNextRow*6 + 4], SequenceData[NextNextRow*6 + 5]),
      Interp).Normalized;
    BenchTicks += 14.28571428571429;
  end;
end;

procedure TGame.Frame;
begin
  if GBenchmark then begin
    CameraOverride:=BenchCam;
    BenchCam.NearPlane:=Scene.Camera.NearPlane;
    BenchCam.FarPlane:=Scene.Camera.FarPlane;
    BenchCam.Aspect:=Scene.Camera.Aspect;
  end;
  inherited Frame;
end;

procedure TGame.PostFrame;
begin
  inherited PostFrame;
  if Assigned(Player) and TPetra(Player).IsDead then
    GRenderer.DrawText(10, VRes - 23,
      {$IFDEF PETRAGCW}
      'Press Y to continue...',
      {$ELSE}
      {$IFDEF ANDROID}
      'Press a button to continue...',
      {$ELSE}
      'Press Space to continue...',
      {$ENDIF}
      {$ENDIF}
      TBitFont(GResourceManager.DefaultFont), 15, 0);
end;

function TGame.KeyDown(Key: TInputKey): Boolean;
begin
  if Key=ikEscape then QuitEngine;
  Result:=True;
end;

function TGame.KeyUp(Key: TInputKey): Boolean;
begin
  Result:=True;
end;

function TGame.Mouse(X, Y: Integer): Boolean;
begin
  Result:=True;
end;

initialization
  RegisterSerializableClass(TGame);
end.
