{ Speech screen }
unit SpchScn;

interface

uses
  InputDrv, Screens, Textures, World;

type
  { TSpeechScreenTheme - The theme to use for the screen box }
  TSpeechScreenTheme = (sstPetra, sstComputer, sstHelen);

  { TSpeechScreen - Screen that appears to indicate spoken words, blabla, etc }
  TSpeechScreen = class(TScreen)
  private
    // Property storage
    FBlankOut: Boolean;
  private
    // Speech border texture
    Texture: TTexture;
    // Texture offset and box position and bgcolor for theme
    TX, TY: Integer;
    BoxPos: Byte;
    // Text to show
    Text: string;
    // Target for post-screen command
    Target: string;
    // Post-screen command to send to the target
    Command: string;
    // The time stamp when the speech started
    StartTime: Integer;
    // Last character to draw count (updated in Draw)
    CharsToDraw: Integer;
  protected
    // TScreen methods
    procedure Draw; override;
    procedure Entered; override;
    procedure Exited; override;
  public
    // TInputHandler methods
    function KeyDown(Key: TInputKey): Boolean; override;
  public
    // TScreen methods
    function Pause: Boolean; override;
    function FullScreen: Boolean; override;
  public
    constructor Create;
    // Activate the speech screen with the given theme and text.  After the
    // user has closed the screen, send the given command to the given entity
    procedure Speech(ATheme: TSpeechScreenTheme; AText, ATarget, ACommand: string);
    // Speech will be drawn blanked out
    property BlankOut: Boolean read FBlankOut write FBlankOut;
  end;

implementation

uses
  GfxDrv, Engine, Game, Fonts, Renderer, Settings;

{ TSpeechScreen }
constructor TSpeechScreen.Create;
begin
  inherited Create;
  // Load speech border texture
  Texture:=TTexture(GResourceManager.GetResource('speech.raw'));
end;

procedure TSpeechScreen.Draw;
var
  I, Y: Integer;
  Line: string[80];
begin
  // Blank out (if requested)
  if BlankOut then GRenderer.FillRectangle(0, 0, HRes, VRes, 0);
  // Draw the speech box
  GRenderer.DrawTexturePart(Texture, BaseX + 16, BaseY + BoxPos, TX, TY, TX + 7, TY + 7);
  GRenderer.DrawTexturePart(Texture, BaseX + 16, BaseY + BoxPos + 60, TX, TY + 16, TX + 7, TY + 23);
  GRenderer.DrawTexturePart(Texture, BaseX + 296, BaseY + BoxPos, TX + 16, TY, TX + 23, TY + 7);
  GRenderer.DrawTexturePart(Texture, BaseX + 296, BaseY + BoxPos + 60, TX + 16, TY + 16, TX + 23, TY + 23);
  I:=BoxPos + 8;
  while I < BoxPos + 60 do begin
    GRenderer.DrawTexturePart(Texture, BaseX + 16, BaseY + I, TX, TY + 8, TX + 7, TY + 15);
    GRenderer.DrawTexturePart(Texture, BaseX + 296, BaseY + I, TX + 16, TY + 8, TX + 23, TY + 15);
    Inc(I, 8);
  end;
  I:=24;
  while I < 296 do begin
    GRenderer.DrawTexturePart(Texture, BaseX + I, BaseY + BoxPos, TX + 8, TY, TX + 15, TY + 7);
    GRenderer.DrawTexturePart(Texture, BaseX + I, BaseY + BoxPos + 60, TX + 8, TY + 16, TX + 15, TY + 23);
    Inc(I, 8);
  end;
  GRenderer.ShadeRectangle(BaseX + 20, BaseY + BoxPos + 4, BaseX + 299, BaseY + BoxPos + 63);
  // Calculate how many characters to draw
  case GSettings.SpeechTextSpeed of
    stsNormal: CharsToDraw:=((GGame.Milliseconds + 100000) - StartTime) div 60;
    stsFast: CharsToDraw:=((GGame.Milliseconds + 100000) - StartTime) div 30;
    stsInstant: CharsToDraw:=Length(Text);
  end;
  if CharsToDraw > Length(Text) then CharsToDraw:=Length(Text);
  // Draw the text
  Y:=BoxPos + 8;
  Line:='';
  for I:=1 to CharsToDraw do begin
    case Text[I] of
      '#': begin end; // Ignore #, it is only used to delay text drawing
      '_': begin // New line
        GRenderer.DrawText(BaseX + 32, BaseY + Y, Line, TBitFont(GResourceManager.DefaultFont), 15, 0);
        Inc(Y, 13);
        Line:='';
      end;
      else Line += Text[I];
    end;
  end;
  GRenderer.DrawText(BaseX + 32, BaseY + Y, Line, TBitFont(GResourceManager.DefaultFont), 15, 0);
  // Draw prompt if the entire text is visible
  if CharsToDraw=Length(Text) then
    GRenderer.DrawText(BaseX + 230, BaseY + BoxPos + 50, 'Continue...', TBitFont(GResourceManager.DefaultFont), 103, 0);
end;

procedure TSpeechScreen.Entered;
begin
  inherited Entered;
  // Save start time stamp (we add 100000 so that if the user presses space,
  // enter or whatever at the first speech screen then setting StartTime to 0
  // above will still work)
  StartTime:=GGame.Milliseconds + 100000;
  // Make the screen manager the exclusive input handler to avoid other handlers
  // getting key events
  GInput.ExclusiveHandler:=GScreenManager;
end;

procedure TSpeechScreen.Exited;
var
  TargetEntity: TEntity;
begin
  // Release the screen manager from being the exclusive handler
  GInput.ExclusiveHandler:=nil;
  // Check if we need to send a command
  if Command <> '' then begin
    TargetEntity:=GGame.World.FindEntity(Target);
    if Assigned(TargetEntity) then TargetEntity.PerformCommand(Command);
  end;
  inherited Exited;
end;

function TSpeechScreen.KeyDown(Key: TInputKey): Boolean;
begin
  if inherited KeyDown(Key) then Exit(True);
  case Key of
    ikControl, ikSpace, ikEnter, ikLeftButton: begin
      // If there are still characters drawn, show them all
      if CharsToDraw < Length(Text) then
        StartTime:=0
      else // ..otherwise exit from the screen
        GScreenManager.Screen:=nil;
    end;
    ikEscape: GScreenManager.Screen:=nil;
  end;
  Result:=False;
end;

function TSpeechScreen.Pause: Boolean;
begin
  Result:=False;
end;

function TSpeechScreen.FullScreen: Boolean;
begin
  Result:=(inherited FullScreen) or BlankOut;
end;

procedure TSpeechScreen.Speech(ATheme: TSpeechScreenTheme; AText, ATarget,
  ACommand: string);
begin
  // Store values
  Text:=AText;
  Target:=ATarget;
  Command:=ACommand;
  // Setup texture offset and bgcolor based on theme
  case ATheme of
    sstPetra: begin
      TX:=0;
      TY:=0;
      BoxPos:=120;
    end;
    sstComputer: begin
      TX:=24;
      TY:=0;
      BoxPos:=40;
    end;
    sstHelen: begin
      TX:=0;
      TY:=24;
      BoxPos:=40;
    end;
  end;
  // Activate the screen
  GScreenManager.Screen:=Self;
end;

end.
